(define boyer-iters        10)
(define browse-iters     1000)
(define conform-iters      20)
(define cpstak-iters     1000)
(define ctak-iters        100)
(define dderiv-iters   800000)
(define deriv-iters    800000)
(define destruc-iters     300)
(define diviter-iters 1000000)
(define divrec-iters  1000000)
(define earley-iters      100)
(define fft-iters        2000)
(define fib-iters          50)
(define fibfp-iters        50)
(define gcold-iters     10000)
(define maze-iters       2500)
(define mazefun-iters    1000)
(define mbrot-iters       100)
(define nucleic-iters      10)
(define parsing-iters    1000)
(define peval-iters       100)
(define pnpoly-iters   100000)
(define puzzle-iters      100)
(define ray-iters          10)
(define reading-iters    1000)
(define scheme-iters    10000)
(define simplex-iters   60000)
(define slatex-iters      100)
(define sum-iters       50000)
(define sumfp-iters     10000)
(define tak-iters        5000)
(define takl-iters        200)
;(define trav1-iters        50) ; non-portable use of quasiquote
;(define trav2-iters        10) ; non-portable use of quasiquote
(define triangl-iters      10)

(define smlboyer-iters     10)
(define nboyer-iters        3) ; problem size, not iterations
(define sboyer-iters        3) ; problem size, not iterations
(define dynamic-iters      50)
(define graphs-iters        1)
(define lattice-iters       1)
;(define nbody-iters         1) ; nondeterministic (order of evaluation)
;(define quicksort-iters    10) ; uses bignums in rand, must fix
(define perm9-iters        10)
