(define boyer-iters    1)
(define browse-iters   1)
(define conform-iters  1)
(define cpstak-iters   1)
(define ctak-iters     1)
(define dderiv-iters   1)
(define deriv-iters    1)
(define destruc-iters  1)
(define diviter-iters  1)
(define divrec-iters   1)
(define earley-iters   1)
(define fft-iters      1)
(define fib-iters      1)
(define fibfp-iters    1)
(define maze-iters     1)
(define mazefun-iters  1)
(define mbrot-iters    1)
(define nucleic-iters  1)
(define peval-iters    1)
(define pnpoly-iters   1)
(define puzzle-iters   1)
(define ray-iters      1)
(define scheme-iters   1)
(define simplex-iters  1)
(define slatex-iters   1)
(define sum-iters      1)
(define sumfp-iters    1)
(define tak-iters      1)
(define takl-iters     1)
(define trav1-iters    1)
(define trav2-iters    1)
(define triangl-iters  1)

(define smlboyer-iters 1)
(define nboyer-iters   0) ; problem size, not iterations
(define dynamic-iters  1)
(define graphs-iters   1)
(define lattice-iters  1)
(define nbody-iters    1)
(define quicksort-iters 1)
(define perm9-iters    1)
