(define boyer-iters        1)
(define browse-iters      60)
(define conform-iters      2)
(define cpstak-iters      30)
(define ctak-iters         3)
(define dderiv-iters   80000)
(define deriv-iters    80000)
(define destruc-iters     30)
(define diviter-iters  40000)
(define divrec-iters   40000)
(define earley-iters      15)
(define fft-iters        200)
(define fib-iters          5)
(define fibfp-iters        5)
(define gcold-iters     1000)
(define maze-iters       250)
(define mazefun-iters     10)
(define mbrot-iters        3)
(define nucleic-iters      1)
(define parsing-iters    100)
(define peval-iters       10)
(define pnpoly-iters    1000)
(define puzzle-iters      10)
(define ray-iters          1)
(define scheme-iters     300)
(define simplex-iters   6000)
(define slatex-iters       2)
(define sum-iters       1000)
(define sumfp-iters     1000)
(define tak-iters        100)
(define takl-iters        20)
(define trav1-iters        5)
(define trav2-iters        1)
(define triangl-iters      1)

(define smlboyer-iters      1)
(define nboyer-iters        0) ; problem size, not iterations
(define sboyer-iters        0) ; problem size, not iterations
(define dynamic-iters       1)
(define graphs-iters        1)
(define lattice-iters       0) ; interpreters take too long on this
;(define nbody-iters         0) ; interpreters take too long on this
;(define quicksort-iters     1)
(define perm9-iters         0) ; interpreters take too long on this
